@@ -1,5 +1,6 @@
 import graphlib
 import logging
+import signal
 import sys
 import time
 from collections import deque
@@ -7,7 +8,8 @@ from concurrent import futures
 from contextvars import Context, ContextVar, copy_context
 from datetime import timedelta
 from subprocess import CalledProcessError
-from typing import Dict, Generator, List, Optional, Tuple
+from types import FrameType
+from typing import Callable, Dict, Generator, List, Optional, Set, Tuple, cast
 
 from colorama import Fore, Style
 
@@ -25,7 +27,7 @@ from ._exceptions import (
 )
 from ._log_capture import PipePlexer
 from ._logging import set_context_handler
-from ._subproc import set_subprocess_default_pipes
+from ._subproc import ProcessManager, set_subprocess_default_pipes
 from ._timing import get_timedelta_since
 
 LOGGER = logging.getLogger(__name__)
@@ -53,52 +55,63 @@ class ExceptionWithTimeSpentException(Exception):
 class Pipeline:
     def __init__(self, config: Config) -> None:
         self.config = config
+        self.proc_manager = ProcessManager()
 
-        self._registered_steps: List[Tuple[str, Step]] = []
+        self._registered_steps: List[Tuple[str, Step, Optional[str]]] = []
         self._registered_step_groups: List[
-            Tuple[str, List[str], Optional[bool]]
+            Tuple[str, List[str], Optional[str], Optional[bool]]
         ] = []
         self._steps_cache: Optional[Dict[str, BaseStepHandler]] = None
 
     @property
-    def _steps(self) -> Dict[str, BaseStepHandler]:
+    def steps(self) -> Dict[str, BaseStepHandler]:
         if self._steps_cache is None:
             self._steps_cache = self._resolve_steps()
         return self._steps_cache
 
-    def register_step(self, name: str, step: "Step") -> None:
-        self._registered_steps.append((name, step))
+    def register_step(
+        self, name: str, description: Optional[str], step: "Step"
+    ) -> None:
+        self._registered_steps.append((name, step, description))
 
     def register_step_group(
         self,
         name: str,
         requires: List[str],
+        description: Optional[str] = None,
         run_by_default: Optional[bool] = None,
     ) -> None:
-        self._registered_step_groups.append((name, requires, run_by_default))
+        self._registered_step_groups.append(
+            (name, requires, description, run_by_default)
+        )
 
     def _resolve_steps(self) -> Dict[str, BaseStepHandler]:
         steps = {}
 
-        for name, func in self._registered_steps:
-            for step in self._resolve_parameters(name, func):
+        for name, func, description in self._registered_steps:
+            for step in self._resolve_parameters(name, func, description):
                 if step.name in steps:
                     raise DuplicateStepException(step.name)
 
                 steps[step.name] = step
 
-        for name, requires, run_by_default in self._registered_step_groups:
+        for (
+            name,
+            requires,
+            description,
+            run_by_default,
+        ) in self._registered_step_groups:
             if name in steps:
                 raise DuplicateStepException(name)
 
             steps[name] = StepGroupHandler(
-                name, self, requires, run_by_default
+                name, self, requires, run_by_default, description
             )
 
         return steps
 
     def _resolve_parameters(
-        self, name: str, func: Step
+        self, name: str, func: Step, description: Optional[str]
     ) -> Generator[BaseStepHandler, None, None]:
         parameters = extract_parameters(func)
         all_run_by_default = True
@@ -106,17 +119,26 @@ class Pipeline:
 
         for params_id, args in parameters:
             step_name = ""
+            current_description = description
             if len(parameters) > 1 and params_id != "":
                 step_name = f"[{params_id}]"
 
             step_name = f"{name}{step_name}"
             current_run_by_default = args.pop("run_by_default", None)
 
+            if "description" in args:
+                current_description = args.pop("description").format(**args)
+            elif description is not None:
+                current_description = description.format(**args)
+            else:
+                current_description = None
+
             all_created.append(step_name)
             all_run_by_default = all_run_by_default and current_run_by_default
 
             yield StepHandler(
                 name=step_name,
+                description=current_description,
                 func=func,
                 pipeline=self,
                 python=args.pop("python", None),
@@ -128,25 +150,45 @@ class Pipeline:
             )
 
         if len(parameters) > 1:
-            yield StepGroupHandler(name, self, all_created, all_run_by_default)
+            yield StepGroupHandler(
+                name, self, all_created, all_run_by_default, description
+            )
 
-    def _resolve_execution_order(
+    def _build_graph(
         self,
         steps: Optional[List[str]] = None,
-        only_steps: Optional[List[str]] = None,
         except_steps: Optional[List[str]] = None,
-    ) -> List[str]:
-        assert not (only_steps and except_steps)
-        if only_steps:
-            steps = only_steps
+        only_selected_steps: bool = False,
+    ) -> Dict[str, List[str]]:
+        # we should refactor at some point
+        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
+
+        # First build the whole graph, without ignoring edges. This is necessary
+        # to ensure we keep all dependency relations
+        def expand(steps: List[str]) -> Set[str]:
+            expanded = set()
+            to_process = deque(steps)
+            while to_process:
+                step = to_process.pop()
+                expanded.add(step)
+                step_info = self.steps[step]
+                if isinstance(step_info, StepGroupHandler):
+                    for req in step_info.requires:
+                        if req not in expanded:
+                            to_process.append(req)
+
+            return expanded
+
         if except_steps is None:
-            except_steps = []
+            except_steps_set = set()
+        else:
+            except_steps_set = expand(except_steps)
 
         if steps is None:
             steps = [
                 name
-                for name, step in self._steps.items()
-                if step.run_by_default and name not in except_steps
+                for name, step in self.steps.items()
+                if step.run_by_default and name not in except_steps_set
             ]
 
         graph = {}
@@ -157,32 +199,94 @@ class Pipeline:
             step = steps_to_process.pop()
 
             try:
-                step_info = self._steps[step]
+                step_info = self.steps[step]
             except KeyError:
                 unknown_steps.append(step)
                 continue
 
-            required_steps = step_info.requires
-            if only_steps:
-                required_steps = [r for r in required_steps if r in only_steps]
-            elif except_steps:
-                required_steps = [
-                    r for r in required_steps if r not in except_steps
-                ]
+            graph[step] = step_info.requires
 
-            graph[step] = required_steps
-
-            for requirement in required_steps:
-                if (
-                    requirement not in graph
-                    and requirement not in except_steps
-                ):
+            for requirement in step_info.requires:
+                if requirement not in graph:
                     steps_to_process.append(requirement)
 
         if unknown_steps:
             raise UnknownStepsException(unknown_steps)
 
+        if except_steps_set:
+            except_replacements: Dict[str, List[str]] = {}
+
+            # FIXME: ensure we handle cycles
+            def compute_replacement(requirements: List[str]) -> List[str]:
+                replacements = []
+                for requirement in requirements:
+                    if requirement in cast(List[str], steps):
+                        replacements.append(requirement)
+                    else:
+                        if requirement not in except_replacements:
+                            except_replacements[
+                                requirement
+                            ] = compute_replacement(graph[requirement])
+                        replacements.extend(except_replacements[requirement])
+
+                return replacements
+
+            for step in except_steps_set:
+                if step not in except_replacements:
+                    # The step might not be in the graph, if it is not depended
+                    # upon by anything else
+                    if deps := graph.get(step):
+                        except_replacements[step] = compute_replacement(deps)
+
+            graph = {
+                key: [
+                    x for v in value for x in except_replacements.get(v, [v])
+                ]
+                for key, value in graph.items()
+                if key not in except_steps_set
+            }
+
+        if only_selected_steps:
+            only = expand(steps)
+            only_replacements = {}
+
+            def compute_only_replacement(
+                step: str, requirements: List[str]
+            ) -> List[str]:
+                if step in only:
+                    return [step]
+
+                replacements = []
+                for requirement in requirements:
+                    if requirement not in only_replacements:
+                        only_replacements[
+                            requirement
+                        ] = compute_only_replacement(
+                            requirement, graph[requirement]
+                        )
+                    replacements.extend(only_replacements[requirement])
+
+                return replacements
+
+            for step in graph.keys():
+                if step not in only_replacements:
+                    only_replacements[step] = compute_only_replacement(
+                        step, graph[step]
+                    )
+
+            graph = {
+                key: [x for v in value for x in only_replacements.get(v, [v])]
+                for key, value in graph.items()
+                if key in only
+            }
+
+        return graph
+
+    def _resolve_execution_order(
+        self, graph: Dict[str, List[str]]
+    ) -> List[str]:
         sorter = graphlib.TopologicalSorter(graph)
+
         try:
             return list(sorter.static_order())
         except graphlib.CycleError as exc:
@@ -191,47 +295,138 @@ class Pipeline:
     def execute(
         self,
         steps: Optional[List[str]],
-        only_steps: Optional[List[str]],
         except_steps: Optional[List[str]],
+        only_selected_steps: bool,
         clean: bool,
     ) -> None:
         # we should refactor at some point
         # pylint: disable=too-many-branches,too-many-locals,too-many-statements
         start_time = time.monotonic()
 
-        steps = self._resolve_execution_order(steps, only_steps, except_steps)
-        if only_steps is None:
-            only_steps = steps
-        if except_steps is None:
-            except_steps = []
+        graph = self._build_graph(steps, except_steps, only_selected_steps)
+        steps_in_order = self._resolve_execution_order(graph)
 
         if clean:
             LOGGER.debug("Cleaning up workspace")
-            for step in steps:
-                self._steps[step].clean()
+            for step in steps_in_order:
+                self.steps[step].clean()
 
-        LOGGER.info("Running steps: %s", ", ".join(steps))
-
-        graph = {
-            step: [
-                r
-                for r in self._steps[step].requires
-                if r in only_steps and r not in except_steps
-            ]
-            for step in steps
-        }
+        LOGGER.info("Running steps: %s", ", ".join(steps_in_order))
 
         sorter = graphlib.TopologicalSorter(graph)
         sorter.prepare()
 
-        results: Dict[str, Tuple[Optional[Exception], timedelta]] = {}
         should_stop = False
+        running_futures: Dict[
+            futures.Future[timedelta], Tuple[str, Optional[PipePlexer]]
+        ] = {}
+
+        def stop() -> None:
+            nonlocal should_stop
+            should_stop = True
+            for future, (name, _) in running_futures.items():
+                LOGGER.debug("Cancelling step %s", name)
+                future.cancel()
+
+        def request_stop(_signum: int, _frame: Optional[FrameType]) -> None:
+            if not should_stop:
+                LOGGER.warning(
+                    "%sStopping requested. This will finish current jobs."
+                    " To abort, hit ^C a second time.",
+                    Style.BRIGHT,
+                )
+                stop()
+            else:
+                LOGGER.critical("Aborting")
+                self.proc_manager.kill()
+
+        previous_signal = signal.signal(signal.SIGINT, request_stop)
+
+        try:
+            results = self._execute(
+                sorter, running_futures, stop, lambda: should_stop
+            )
+        finally:
+            signal.signal(signal.SIGINT, previous_signal)
+
+        self._log_summary(graph, results, start_time)
+
+    def get_step(self, step_name: str) -> BaseStepHandler:
+        return self.steps[step_name]
+
+    def list_all_steps(
+        self,
+        steps: Optional[List[str]] = None,
+        except_steps: Optional[List[str]] = None,
+        only_selected_steps: bool = False,
+        show_dependencies: bool = False,
+    ) -> None:
+        # pylint: disable=too-many-locals
+        all_steps = self._resolve_execution_order(
+            self._build_graph(list(self.steps.keys()))
+        )
+        selected_steps = self._resolve_execution_order(
+            self._build_graph(steps, except_steps, only_selected_steps)
+        )
+
+        step_infos = []
+        for step in sorted(all_steps):
+            step_info = self.steps[step]
+
+            dep_info = ""
+            if show_dependencies and step_info.requires:
+                dep_info = " --> " + ", ".join(
+                    reversed([s for s in all_steps if s in step_info.requires])
+                )
+
+            description = ""
+            if self.config.verbosity > 0 and step_info.description:
+                description = step_info.description
+
+            step_infos.append((step, dep_info, description))
+
+        LOGGER.info("Available steps (* means selected, - means skipped):")
+        if not step_infos:
+            return
+
+        max_step_length = max(len(s[0]) for s in step_infos)
+        max_dependencies_length = max(len(s[1]) for s in step_infos)
+
+        for step, dependencies, description in step_infos:
+            indicator = "*"
+            style = Style.BRIGHT
+            if step not in selected_steps:
+                style = ""
+                indicator = "-"
+
+            if self.config.verbosity > 0 and description:
+                description = f"\t[{Fore.BLUE}{Style.NORMAL}{description}{Style.RESET_ALL}{style}]"
+            else:
+                description = ""
+
+            LOGGER.info(
+                "\t%s%s %-*s%-*s%s",
+                style,
+                indicator,
+                max_step_length,
+                step,
+                max_dependencies_length,
+                dependencies,
+                description,
+            )
+
+    def _execute(
+        self,
+        sorter: "graphlib.TopologicalSorter[str]",
+        running_futures: Dict[
+            futures.Future[timedelta], Tuple[str, Optional[PipePlexer]]
+        ],
+        stop: Callable[[], None],
+        should_stop: Callable[[], bool],
+    ) -> Dict[str, Tuple[Optional[Exception], timedelta]]:
+        results: Dict[str, Tuple[Optional[Exception], timedelta]] = {}
 
         with futures.ThreadPoolExecutor(self.config.n_jobs) as executor:
-            running_futures: Dict[
-                futures.Future[timedelta], Tuple[str, Optional[PipePlexer]]
-            ] = {}
-
             while sorter.is_active():
                 ready = sorter.get_ready()
                 if not ready:
@@ -263,9 +458,7 @@ class Pipeline:
                             sorter.done(name)
                         else:
                             if self.config.fail_fast:
-                                should_stop = True
-                                for future in running_futures:
-                                    future.cancel()
+                                stop()
 
                             # We won't be able to enqueue new results after a failure
                             # anyways
@@ -277,7 +470,7 @@ class Pipeline:
                         results[name] = None, time_spent
                         sorter.done(name)
 
-                if should_stop:
+                if should_stop():
                     continue
 
                 for name in ready:
@@ -298,40 +491,7 @@ class Pipeline:
                     )
                     running_futures[future] = name, pipe_plexer
 
-            self._log_summary(graph, results, start_time)
-
-    def get_step(self, step_name: str) -> BaseStepHandler:
-        return self._steps[step_name]
-
-    def list_all_steps(
-        self,
-        steps: Optional[List[str]] = None,
-        only_steps: Optional[List[str]] = None,
-        except_steps: Optional[List[str]] = None,
-        show_dependencies: bool = False,
-    ) -> None:
-        all_steps = self._resolve_execution_order(list(self._steps.keys()))
-        selected_steps = self._resolve_execution_order(
-            steps, only_steps, except_steps
-        )
-
-        LOGGER.info("Available steps (* means selected, - means skipped):")
-        for step in sorted(all_steps):
-            dep_info = ""
-            if show_dependencies and self._steps[step].requires:
-                dep_info = " --> " + ", ".join(
-                    reversed(
-                        [
-                            s
-                            for s in all_steps
-                            if s in self._steps[step].requires
-                        ]
-                    )
-                )
-            if step in selected_steps:
-                LOGGER.info("\t* %s%s", step, dep_info)
-            else:
-                LOGGER.info("\t%s- %s%s", Fore.LIGHTBLACK_EX, step, dep_info)
+        return results
 
     def _log_summary(
         self,
@@ -351,23 +511,32 @@ class Pipeline:
                 result, time_spent = results[name]
                 if result is None:
                     LOGGER.info(
-                        "\t%s[%s] %s: success", Fore.GREEN, time_spent, name
+                        "\t%s[%s] %s%s%s: success",
+                        Fore.GREEN,
+                        time_spent,
+                        Style.BRIGHT,
+                        name,
+                        Style.NORMAL,
                     )
                 elif (
                     isinstance(result, UnavailableInterpreterException)
                     and self.config.skip_missing_interpreters
                 ):
                     LOGGER.info(
-                        "\t%s%s: Skipped: %s",
+                        "\t%s[-:--:--.------] %s%s%s: Skipped: %s",
                         Fore.YELLOW,
+                        Style.BRIGHT,
                         name,
+                        Style.NORMAL,
                         result,
                     )
                 elif isinstance(result, futures.CancelledError):
                     LOGGER.info(
-                        "\t%s%s: Cancelled",
+                        "\t%s[-:--:--.------] %s%s%s: Cancelled",
                         Fore.YELLOW,
+                        Style.BRIGHT,
                         name,
+                        Style.NORMAL,
                     )
                     cancelled_jobs.append(name)
                 else:
@@ -382,9 +551,11 @@ class Pipeline:
                     failed_jobs.append(name)
             elif cancelled_jobs:
                 LOGGER.info(
-                    "\t%s%s: Cancelled",
+                    "\t%s[-:--:--.------] %s%s%s: Cancelled",
                     Fore.YELLOW,
+                    Style.BRIGHT,
                     name,
+                    Style.NORMAL,
                 )
                 cancelled_jobs.append(name)
             else:
@@ -396,9 +567,11 @@ class Pipeline:
                 assert blocking_dependencies is not None
                 blocked_jobs.append(name)
                 LOGGER.warning(
-                    "\t%s%s: blocked by %s",
+                    "\t%s[-:--:--.------] %s%s%s: blocked by %s",
                     Fore.YELLOW,
+                    Style.BRIGHT,
                     name,
+                    Style.NORMAL,
                     ", ".join(blocking_dependencies),
                 )
 
@@ -445,9 +618,11 @@ class Pipeline:
         start_time = time.monotonic()
 
         try:
-            self._steps[name].execute()
+            self.steps[name].execute()
         except UnavailableInterpreterException as exc:
-            LOGGER.warning("Step %s failed: %s", name, exc)
+            LOGGER.warning(
+                "Step %s%s%s failed: %s", Style.BRIGHT, name, Style.NORMAL, exc
+            )
             raise ExceptionWithTimeSpentException(
                 exc, get_timedelta_since(start_time)
             ) from exc
@@ -464,7 +639,13 @@ class Pipeline:
                 exc, get_timedelta_since(start_time)
             ) from exc
 
-        LOGGER.info("%sStep %s finished successfully", Fore.GREEN, name)
+        LOGGER.info(
+            "%sStep %s%s%s finished successfully",
+            Fore.GREEN,
+            Style.BRIGHT,
+            name,
+            Style.NORMAL,
+        )
         return get_timedelta_since(start_time)
 
     def _format_exception(self, exc: Exception) -> str:
